/*
Module : PJNSNTP.cpp
Purpose: Implementation for a MFC class to encapsulate the SNTP protocol
Created: PJN / 05-08-1998
History: PJN / 16-11-1998 1. m_nOriginateTime was getting set incorrectly in the SNTP response
                          2. GetLastError now works when a timeout occurs.
         PJN / 25-06-2000 1. Fixed an issue where the packing order of the NtpBasicInfo
                          struct was causing problems. Now packing is set to 1 byte for this
                          struct
         AR  / 22-07-2000 Windows CE 2.11 compatibility
         PJN / 29-03-2001 1. Updated copyright message
                          2. Fixed a handle leak RevertSetTimePriviledge where the process
                          token was not been closed
         PJN / 13-03-2003 1. Fixed a bug in CNtpSocket::IsReadible in the setup of the timeval
                          structure. Thanks to "InBloom Support" for reporting this.
         PJN / 20-09-2003 1. Class now uses CWSocket class to provide sockets capability. This
                          now allows the class to support connecting via a Socks 5 proxy.
         PJN / 17-03-2005 1. Updated to work with the latest release of CWSocket. Thanks to 
                          Alex Dron for reporting this issue.
                          2. Addition of a preprocessor macro "PJNSNTP_EXT_CLASS" which allows
                          the code to be easily added to an extension DLL.
                          3. General tidy up of the code in CPJNSNTPClient::GetServerTime
                          4. Replaced the macro _WIN32_WCE with PJNSNTP_NO_PRIVILEGE_CODE. This 
                          allows the priviledge code to also be excluded on desktop versions of
                          Windows if so desired.  
                          5. Reviewed and updated all the TRACE statements for correctness.
         PJN / 04-04-2005 1. Fix for a bug in the calculation of the m_RoundTripDelay value. Thanks
                          to Euan Lee for reporting this bug.
         PJN / 10-07-2005 1. Undid the change made on 04-04-2005 as the calculation of the round 
                          trip time calculation was correct. For details on the formulae used,
                          please check out RFC 2030. Thanks to Thomas Fejes for reporting this
                          problem.
         PJN / 12-09-2005 1. Fixed problem where code did not handle graceful disconnects when 
                          reading the SNTP response.
                          2. Reworked the class to now pass exceptions rather than use a Win32 
                          style return value.
                          3. Renamed the classes to use a prefix of PJN. This is in line with the
                          SMTP class I also provide.
         PJN / 05-06-2006 1. Updated copyright details.
                          2. Updated the documentation to use the same style as the web site.
                          3. Code now uses newer C++ style casts instead of C style casts.
                          4. Optimized CPJNNTPTime constructor code.
                          5. Optimized CPJNSNTPClient constructor code
                          6. Replaced calls to ZeroMemory with memset
                          7. Updated code to work with latest version of CWSocket class.
                          8. Made the code more robust in its handling of the m_hToken variable
                          9. Optimized the privilege handling code to avoid the need for the 
                          m_bTakenPrivilege variable.
                          10. Verified all code compiles cleanly on VC 2005
                          11. Renamed PJNSNTP_NO_PRIVILEDGE_CODE macro to PJNSNTP_NO_PRIVILEGE_CODE
         PJN / 11-07-2008 1. Updated copyright details.
                          2. Updated sample app to clean compile on VC 2008
                          3. Removed VC 6 style classwizard comments from the code.
                          4. The code has now been updated to support VC 2005 or later only. 
                          5. Code now compiles cleanly using Code Analysis (/analyze)
         PJN / 16-03-2009 1. Updated copyright details.
                          2. Included missing VC 2005 solution and project files in the download. Thanks 
                          to Michael Haephrati for reporting this issue.
         PJN / 06-04-2015 1. Updated the code to clean compile on VC 2010 and later
                          2. Updated the code to compile with the latest version of the CWSocket class
                          3. Removed all the code which refers to a proxy connection
                          4. Updated the code to operate optionally independent of MFC. If the preprocessor
                          value "CWSOCKET_MFC_EXTENSTIONS" is not defined then the SNTP code will not
                          use MFC, while if it is defined then the code will use MFC. This preprocessor
                          value is used to control the MFC mode of the CWSocket class.
                          5. CPJNSNTPClient::SetClientTime now preserves the last error value and the 
                          sample app now reports this value.
                          6. Optimized the implementation of CPJNNTPTime::MsToNtpFraction, NtpFractionToMs
                          & NtpFractionToSecond methods. These optimizations have got rid of the the need
                          for the previous m_MsToNTP lookup table.
                          7. Reworked the CPJNNTPTime::CPJNNTPTime(const SYSTEMTIME& st) constructor to
                          use SystemTimeToFileTime Windows API
                          8. Reworked the CPJNNTPTime::operator SYSTEMTIME() method to use 
                          FileTimeToSystemTime Windows API
                          9. The sample app now displays the stratum and leap indicator values from the 
                          received SNTP response.
         PJN / 18-01-2016 1. Updated copyright details.
                          2. Verified the code compiles cleanly in VC 2015
                          3. Added SAL annotations to all the code
         PJN / 19-08-2018 1. Updated copyright details.
                          2. Fixed a number of C++ core guidelines compiler warnings. These changes mean 
                          that the code will now only compile on VC 2017 or later.
                          3. Removed PJNSNTP_NO_PRIVILEGE_CODE preprocessor from codebase
         PJN / 01-06-2019 1. Updated copyright details.
                          2. Updated the code to clean compile on VC 2019
         PJN / 16-09-2019 1. Fixed a number of compiler warnings when the code is compiled with VS 2019 Preview
         PJN / 28-10-2019 1. Updated the code to correctly handle SNTP timestamps greater than the first SNTP
                          epoch date of 6:28:16AM on 7 February 2036. Thanks to Mai Duc Hung for reporting this
                          issue.
                          2. Replaced BOOL with bool throughout the code.
         PJN / 30-10-2019 1. Fixed a compiler warning in the EnableSetTimePrivilege method. Thanks to Mai Duc 
                          Hung for reporting this issue.
                          2. Changed macro references of "CWSOCKET_MFC_EXTENSTIONS" to 
                          "CWSOCKET_MFC_EXTENSIONS". Thanks to Mai Duc Hung for reporting this issue.
         PJN / 20-12-2019 1. Updated copyright details.
                          2. Fixed various Clang-Tidy static code analysis warnings in the code.
         PJN / 03-05-2020 1. Fixed more Clang-Tidy static code analysis warnings in the code.
         PJN / 27-04-2022 1. Updated copyright details.
                          2. Updated the code to use C++ uniform initialization for all variable declarations

Copyright (c) 1998 - 2022 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code.

*/


//////////////////// Includes /////////////////////////////////////////////////

#include "stdafx.h"
#include "PJNSNTP.h"


//////////////////// Macros / Locals //////////////////////////////////////////

#ifdef CWSOCKET_MFC_EXTENSIONS
#ifdef _DEBUG
#define new DEBUG_NEW
#endif //#ifdef _DEBUG
#endif //#ifdef CWSOCKET_MFC_EXTENSIONS

//The mandatory part of an NTP packet
#pragma pack(push, 1)
struct NtpBasicInfo
{
  BYTE m_LiVnMode;
  BYTE m_Stratum;
  char m_Poll;
  char m_Precision;
  long m_RootDelay;
  long m_RootDispersion;
  char m_ReferenceID[4]; //NOLINT(modernize-avoid-c-arrays)
  CPJNNTPTimePacket m_ReferenceTimestamp;
  CPJNNTPTimePacket m_OriginateTimestamp;
  CPJNNTPTimePacket m_ReceiveTimestamp;
  CPJNNTPTimePacket m_TransmitTimestamp;
};
#pragma pack(pop)

//The optional part of an NTP packet
struct NtpAuthenticationInfo
{
  unsigned long m_KeyID;
  BYTE m_MessageDigest[16]; //NOLINT(modernize-avoid-c-arrays)
};

//The Full NTP packet
struct NtpFullPacket
{
  NtpBasicInfo m_Basic;
  NtpAuthenticationInfo m_Auth;
};


//////////////////// Implementation ///////////////////////////////////////////

CPJNSNTPClient::CPJNSNTPClient() noexcept : m_dwTimeout{5000},
                                            m_hToken{nullptr},
                                            m_TokenPriv{}
{
}

CPJNSNTPClient::~CPJNSNTPClient()
{
  if (m_hToken != nullptr)
    RevertSetTimePrivilege();
}

void CPJNSNTPClient::GetServerTime(_In_z_ LPCTSTR pszHostName, _Inout_ CPJNNTPServerResponse& response, _In_ int nPort)
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(pszHostName != nullptr);

  //Connect to the SNTP server
  CWSocket socket;
  socket.SetBindAddress(m_sLocalBoundAddress);
  socket.CreateAndConnect(pszHostName, nPort, SOCK_DGRAM);

  //Initialise the NtpBasicInfo packet
  NtpBasicInfo nbi{};
  constexpr const int nSendSize{sizeof(NtpBasicInfo)};
  nbi.m_LiVnMode = 27; //Encoded representation which represents NTP Client Request & NTP version 3.0 (i.e. LI=0, VN=3, Mode=3)
  nbi.m_TransmitTimestamp = CPJNNTPTime::GetCurrentTime();

  //Send off the NtpBasicInfo packet
  socket.Send(&nbi, nSendSize);

  //Need to use select to determine readability of socket
  if (!socket.IsReadable(m_dwTimeout))
    CWSocket::ThrowWSocketException(WSAETIMEDOUT);
  response.m_DestinationTime = CPJNNTPTime::GetCurrentTime();

  //read back the response into the NtpFullPacket struct
  NtpFullPacket nfp{};
  constexpr const int nReceiveSize{sizeof(nfp)};
  if (socket.Receive(&nfp, nReceiveSize) == 0)
    CWSocket::ThrowWSocketException(ERROR_GRACEFUL_DISCONNECT);

  //Transfer all the useful info into the response structure
  response.m_nStratum = nfp.m_Basic.m_Stratum;
  response.m_nLeapIndicator = (nfp.m_Basic.m_LiVnMode & 0xC0) >> 6;
  response.m_OriginateTime = nfp.m_Basic.m_OriginateTimestamp;
  response.m_ReceiveTime = nfp.m_Basic.m_ReceiveTimestamp;
  response.m_TransmitTime = nfp.m_Basic.m_TransmitTimestamp;
  response.m_RoundTripDelay = (response.m_DestinationTime.SecondsAsDouble() - response.m_OriginateTime.SecondsAsDouble()) - (response.m_ReceiveTime.SecondsAsDouble() - response.m_TransmitTime.SecondsAsDouble());
  response.m_LocalClockOffset = ((response.m_ReceiveTime.SecondsAsDouble() - response.m_OriginateTime.SecondsAsDouble()) + (response.m_TransmitTime.SecondsAsDouble() - response.m_DestinationTime.SecondsAsDouble())) / 2;
}

bool CPJNSNTPClient::EnableSetTimePrivilege() noexcept
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(m_hToken == nullptr);

  if (!OpenProcessToken(GetCurrentProcess(), TOKEN_ADJUST_PRIVILEGES | TOKEN_QUERY, &m_hToken))
    return false;
  m_TokenPriv = {};
  if (!LookupPrivilegeValue(nullptr, SE_SYSTEMTIME_NAME, &m_TokenPriv.Privileges[0].Luid))
  {
    //Close the token handle since we have no need for it
    CloseHandle(m_hToken);
    m_hToken = nullptr;

    return false;
  }
  m_TokenPriv.PrivilegeCount = 1;
  m_TokenPriv.Privileges[0].Attributes |= SE_PRIVILEGE_ENABLED;
#pragma warning(suppress: 26477)
  const BOOL bSuccess{AdjustTokenPrivileges(m_hToken, FALSE, &m_TokenPriv, 0, nullptr, nullptr)};
  if (!bSuccess)
  {
    //Close the token handle since we have no need for it
    CloseHandle(m_hToken);
    m_hToken = nullptr;
  }

  return (bSuccess != 0);
}

bool CPJNSNTPClient::RevertSetTimePrivilege() noexcept
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSUME(m_hToken != nullptr);

  m_TokenPriv.Privileges[0].Attributes &= (~SE_PRIVILEGE_ENABLED);
#pragma warning(suppress: 26477)
  const BOOL bSuccess{AdjustTokenPrivileges(m_hToken, FALSE, &m_TokenPriv, 0, nullptr, nullptr)};

  //Free up the token now that we are finished with it
  CloseHandle(m_hToken);
  m_hToken = nullptr;

  return bSuccess != 0;
}

bool CPJNSNTPClient::SetClientTime(_In_ const SYSTEMTIME& st, bool bEnableSetTimePrivilege) noexcept
{
  if (bEnableSetTimePrivilege)
  {
    if (!EnableSetTimePrivilege())
      return false;
  }
  const BOOL bSuccess{SetSystemTime(&st)};
  const DWORD dwLastError{GetLastError()};
  if (bEnableSetTimePrivilege)
    RevertSetTimePrivilege();
  if (!bSuccess)
    SetLastError(dwLastError);
  return bSuccess != 0;
}
