/*
Module : PJNSNTP.h
Purpose: Interface for a C++ class to encapsulate the SNTP protocol
Created: PJN / 05-08-1998

Copyright (c) 1998 - 2022 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code.

*/


//////////////////// Includes / Defines ///////////////////////////////////////

#pragma once

#ifndef __PJNSNTP_H__
#define __PJNSNTP_H__

#ifndef PJNSNTP_EXT_CLASS
#define PJNSNTP_EXT_CLASS
#endif //#ifndef PJNSNTP_EXT_CLASS

#include "SocMFC.h" //If you get a compilation error about this missing header file, then you need to download my CWSocket wrapper classes from http://www.naughter.com/w3mfc.html

#ifndef _BITSET_
#pragma message("To avoid this message, please put bitset in your pre compiled header (usually stdafx.h)")
#include <bitset>
#endif //#ifndef _BITSET_

//////////////////// Classes //////////////////////////////////////////////////

//On the wire representation of an NTP timestamp
struct PJNSNTP_EXT_CLASS CPJNNTPTimePacket
{
  DWORD m_dwInteger;
  DWORD m_dwFractional;
};

//Helper class to encapsulate NTP time stamps
class PJNSNTP_EXT_CLASS CPJNNTPTime
{
public:
//Constructors / Destructors
  CPJNNTPTime() noexcept : m_Time{0}
  {
  }

  CPJNNTPTime(_In_ const CPJNNTPTime& time) noexcept = default;
  CPJNNTPTime(_In_ CPJNNTPTime&& time) noexcept = default;

  CPJNNTPTime(_In_ const CPJNNTPTimePacket& packet) noexcept
  {
    const DWORD dwLow{ntohl(packet.m_dwFractional)};
    const DWORD dwHigh{ntohl(packet.m_dwInteger)};
#pragma warning(suppress: 26472)
    m_Time = ((static_cast<unsigned __int64>(dwHigh)) << 32) + dwLow;
  }

  CPJNNTPTime(_In_ const SYSTEMTIME& st) noexcept
  {
    FILETIME ft{};
    if (SystemTimeToFileTime(&st, &ft))
    {
      constexpr const unsigned __int64 dwEpoch1900{94354848000000000}; //1900-1-1 expressed as a FILETIME
      constexpr const unsigned __int64 dwEpoch2036{137304520960000000}; //2036-2-7 6:28:16 expressed as a FILETIME
#pragma warning(suppress: 26472)
      const unsigned __int64 dwInput{(static_cast<unsigned __int64>(ft.dwHighDateTime) << 32) + ft.dwLowDateTime};
      unsigned __int64 dwSeconds{0};
      if (dwInput >= dwEpoch2036)
        dwSeconds = (dwInput - dwEpoch2036) / 10000000;
      else
        dwSeconds = (dwInput - dwEpoch1900) / 10000000;
      m_Time = (dwSeconds << 32) + MsToNtpFraction(st.wMilliseconds);
    }
    else
      m_Time = 0;
  }

//General functions
  CPJNNTPTime& operator=(_In_ const CPJNNTPTime& time) noexcept = default;

  CPJNNTPTime& operator=(_In_ CPJNNTPTime&& time) noexcept = default;

  _NODISCARD SYSTEMTIME AsSYSTEMTIME() const
  {
    const DWORD dwSeconds{Seconds()};
    const DWORD dwMilliseconds{Milliseconds()};

    unsigned __int64 dwFT{0};
    if (!std::bitset<32>(dwSeconds).test(31)) //If the high bit is not set, then treat the value as after Epoch 2036
      dwFT = 137304520960000000 + (static_cast<unsigned __int64>(dwSeconds) * 10000000) + (static_cast<unsigned __int64>(dwMilliseconds) * 10000);
    else
#pragma warning(suppress: 26472)
      dwFT = 94354848000000000 + (static_cast<unsigned __int64>(dwSeconds) * 10000000) + (static_cast<unsigned __int64>(dwMilliseconds) * 10000);
    FILETIME ft{};
#pragma warning(suppress: 26472)
    ft.dwLowDateTime = static_cast<DWORD>(dwFT & 0x00000000FFFFFFFF);
#pragma warning(suppress: 26472)
    ft.dwHighDateTime = static_cast<DWORD>((dwFT & 0xFFFFFFFF00000000) >> 32);
    SYSTEMTIME st{};
    FileTimeToSystemTime(&ft, &st);
#pragma warning(suppress: 6102)
    return st;
  }

  _NODISCARD FILETIME AsFILETIME() const
  {
    const SYSTEMTIME st{AsSYSTEMTIME()};
    FILETIME ft{};
    SystemTimeToFileTime(&st, &ft);
    return ft;
  }

  operator CPJNNTPTimePacket() const noexcept
  {
    CPJNNTPTimePacket ntp;
    ntp.m_dwInteger = htonl(Seconds());
    ntp.m_dwFractional = htonl(Fraction());
    return ntp;
  }

  operator unsigned __int64() const noexcept
  {
    return m_Time;
  }

  _NODISCARD DWORD Seconds() const noexcept
  {
#pragma warning(suppress: 26472)
    return static_cast<DWORD>((m_Time & 0xFFFFFFFF00000000) >> 32);
  }

  _NODISCARD double SecondsAsDouble() const
  {
    const FILETIME ft{AsFILETIME()};
    return ((static_cast<unsigned __int64>(ft.dwHighDateTime) << 32) + static_cast<unsigned __int64>(ft.dwLowDateTime)) / 10000000.0;
  }

  _NODISCARD DWORD Milliseconds() const noexcept
  {
    return NtpFractionToMs(Fraction());
  }

  _NODISCARD DWORD Fraction() const noexcept
  {
#pragma warning(suppress: 26472)
    return static_cast<DWORD>(m_Time & 0xFFFFFFFF);
  }

//Static functions
  static CPJNNTPTime GetCurrentTime() noexcept
  {
    SYSTEMTIME st{};
    GetSystemTime(&st);
    return {st};
  }

  constexpr static DWORD MsToNtpFraction(_In_ WORD wMilliSeconds) noexcept
  {
#pragma warning(suppress: 26467)
    return static_cast<DWORD>((wMilliSeconds / 1000.0 * 0xFFFFFFFF) + 0.5);
  }

  constexpr static WORD NtpFractionToMs(_In_ DWORD dwFraction) noexcept
  {
#pragma warning(suppress: 26467)
    return static_cast<WORD>(static_cast<double>(dwFraction) * 1000.0 / 0xFFFFFFFF);
  }

protected:
//The actual data
  unsigned __int64 m_Time;
};

class PJNSNTP_EXT_CLASS CPJNNTPServerResponse
{
public:
//Constructors / Destructors
  CPJNNTPServerResponse() noexcept : m_nLeapIndicator{0},
                                     m_nStratum{0},
                                     m_RoundTripDelay{0},
                                     m_LocalClockOffset{0}
  {
  }
  CPJNNTPServerResponse(const CPJNNTPServerResponse&) = default;
  CPJNNTPServerResponse(CPJNNTPServerResponse&&) = default;
  ~CPJNNTPServerResponse() = default;

//Methods
  CPJNNTPServerResponse& operator=(const CPJNNTPServerResponse&) = default;
  CPJNNTPServerResponse& operator=(CPJNNTPServerResponse&&) = default;

//Member variables
  int m_nLeapIndicator; //0: no warning
                        //1: last minute in day has 61 seconds
                        //2: last minute has 59 seconds
                        //3: clock not synchronized
  int m_nStratum; //0: unspecified or unavailable
                  //1: primary reference (e.g., radio clock)
                  //2-15: secondary reference (via NTP or SNTP)
                  //16-255: reserved
  CPJNNTPTime m_OriginateTime; //Time when the request was sent from the client to the SNTP server
  CPJNNTPTime m_ReceiveTime; //Time when the request was received by the server
  CPJNNTPTime m_TransmitTime; //Time when the server sent the request back to the client
  CPJNNTPTime m_DestinationTime; //Time when the reply was received by the client
  double m_RoundTripDelay; //Round trip time in seconds
  double m_LocalClockOffset; //Local clock offset relative to the server
};

//The actual SNTP class
class PJNSNTP_EXT_CLASS CPJNSNTPClient
{
public:
//Constructors / Destructors
  CPJNSNTPClient() noexcept;
  ~CPJNSNTPClient();
  CPJNSNTPClient(const CPJNSNTPClient&) = delete;
  CPJNSNTPClient(CPJNSNTPClient&&) = delete;

//General functions
  CPJNSNTPClient& operator=(const CPJNSNTPClient&) = delete;
  CPJNSNTPClient& operator=(CPJNSNTPClient&&) = delete;
  void GetServerTime(_In_z_ LPCTSTR pszHostName, _Inout_ CPJNNTPServerResponse& response, _In_ int nPort = 123);
  _NODISCARD DWORD GetTimeout() const noexcept { return m_dwTimeout; };
  void SetTimeout(_In_ DWORD dwTimeout) noexcept { m_dwTimeout = dwTimeout; };
  bool SetClientTime(_In_ const SYSTEMTIME& st, bool bEnableSetTimePrivilege = true) noexcept;
  void SetBoundAddress(_In_ const CWSocket::String& sLocalBoundAddress) { m_sLocalBoundAddress = sLocalBoundAddress; };
  _NODISCARD CWSocket::String SetBoundAddress() const { return m_sLocalBoundAddress; };

protected:
//Methods
  bool EnableSetTimePrivilege() noexcept;
  bool RevertSetTimePrivilege() noexcept;

//Member variables
  DWORD m_dwTimeout;
  CWSocket::String m_sLocalBoundAddress;
#ifndef PJNSNTP_NO_PRIVILEGE_CODE
  HANDLE m_hToken;
  TOKEN_PRIVILEGES m_TokenPriv;
#endif
};

#endif //#ifndef __PJNSNTP_H__
